#!/usr/bin/env python
# -*- coding: utf8 -*-

"""
Checks whether a FoLiA document is a valid FoLiA document, i.e. whether it properly adheres to the FoLiA specification. Invalid documents should never be used or published. This tool is one of the most indispensable tools in the toolbox for anyone dealing with FoLiA.
"""

import sys
import os
import glob
import traceback
import lxml.etree
import argparse
from foliatools import VERSION as TOOLVERSION
import folia.main as folia


def validate(filename, schema = None,**kwargs):
    if not kwargs.get('quick'):
        try:
            folia.validate(filename, schema)
        except Exception as e:
            print("VALIDATION ERROR against RelaxNG schema (stage 1/3), in " + filename,file=sys.stderr)
            print(str(e), file=sys.stderr)
            return False
    try:
        document = folia.Document(file=filename, deepvalidation=kwargs.get('deep',False),textvalidation=kwargs.get('stricttextvalidation',False),verbose=True, autodeclare=kwargs.get('autodeclare',False), processor=kwargs.get('processor'), keepversion=kwargs.get('keepversion'), fixunassignedprocessor=kwargs.get('fixunassignedprocessor'), debug=kwargs.get('debug',0))
    except folia.DeepValidationError as e:
        print("DEEP VALIDATION ERROR on full parse by library (stage 2/3), in " + filename,file=sys.stderr)
        print(e.__class__.__name__ + ": " + str(e),file=sys.stderr)
        return False
    except Exception as e:
        print("VALIDATION ERROR on full parse by library (stage 2/3), in " + filename,file=sys.stderr)
        print(e.__class__.__name__ + ": " + str(e),file=sys.stderr)
        if kwargs.get('traceback') or kwargs.get('debug'):
            print("-- Full traceback follows -->",file=sys.stderr)
            ex_type, ex, tb = sys.exc_info()
            traceback.print_exception(ex_type, ex, tb)
        return False
    if not document.version:
        print("VALIDATION ERROR: Document does not advertise FoLiA version (" + filename + ")",file=sys.stderr)
        return False
    elif folia.checkversion(document.version) == -1 and not kwargs.get('nowarn'):
        if document.version.split('.')[0] in ('0','1'):
            print("WARNING: Document (" + filename + ") uses an older FoLiA version ("+document.version+") but is validated with a newer library (" + folia.FOLIAVERSION+"). If this is a document you created and intend to publish, you may want to upgrade this FoLiA v1 document to FoLiA v2 using the 'foliaupgrade' tool.",file=sys.stderr)
        else:
            print("WARNING: Document (" + filename + ") uses an older FoLiA version ("+document.version+") but is validated according to the newer specification (" + folia.FOLIAVERSION+"). You might want to increase the version attribute if this is a document you created and intend to publish.",file=sys.stderr)
    if document.textvalidationerrors:
        if kwargs.get('stricttextvalidation'):
            print("VALIDATION ERROR because of text validation errors (stage 2/3), in " + filename,file=sys.stderr)
            return False
        elif not kwargs.get('nowarn'):
            print("WARNING: there were " + str(document.textvalidationerrors) + " text validation errors but these are currently not counted toward the full validation result (use -t for strict text validation)", file=sys.stderr)
    if not kwargs.get('quick'):
        try:
            if kwargs.get('output'):
                if not (folia.checkversion(document.version, "2.0.0") < 0 and kwargs.get('keepversion')):
                    document.provenance.append( folia.Processor.create(name="foliavalidator", version=TOOLVERSION, src="https://github.com/proycon/foliatools", metadata={"valid": "yes"}) )
            xml = document.xmlstring(form=folia.Form.EXPLICIT if kwargs.get('explicit') else folia.Form.NORMAL)
            if kwargs.get('output'):
                if folia.checkversion(document.version, "2.0.0") < 0 and not kwargs.get('keepversion') and not kwargs.get('autodeclare'):
                    print("WARNING: Document is valid but can't output older FoLiA (< v2) document unless you specify either --keepversion or --autodeclare to attempt to upgrade. However, if you really want to upgrade the document, use the 'foliaupgrade' tool instead." , file=sys.stderr)
                    return False
                else:
                    print(xml)
        except Exception as e:
            print("SERIALISATION ERROR (stage 3/3): Document validated succesfully but failed to serialise! (" + filename + "). This may be indicative of a problem in the underlying library, please submit an issue on https://github.com/proycon/foliapy with the output of this error.",file=sys.stderr)
            print(e.__class__.__name__ + ": " + str(e),file=sys.stderr)
            print("-- Full traceback follows -->",file=sys.stderr)
            ex_type, ex, tb = sys.exc_info()
            traceback.print_exception(ex_type, ex, tb)
            return False
    if kwargs.get('autodeclare'):
        print("Validated successfully **after** applying auto-declarations: " +  filename + " (this is not a guarantee that the original file is valid but indicates it can be automatically made valid!)", file=sys.stderr)
        return document
    else:
        print("Validated successfully: " +  filename,file=sys.stderr)
        return True




def processdir(d, schema = None, **kwargs):
    success = False
    print("Searching in  " + d,file=sys.stderr)
    extension = kwargs.get('extension','xml').strip('.')
    for f in glob.glob(os.path.join(d ,'*')):
        r = True
        if f[-len(extension) - 1:] == '.' + extension:
            r = validate(f, schema,**kwargs)
        elif kwargs.get('recurse') and os.path.isdir(f):
            r = processdir(f,schema,**kwargs)
        if not r: success = False
    return success

def commandparser(parser):
    parser.add_argument('-d','--deep',help="Enable deep validation; validated uses classes against provided set definitions", action='store_true', default=False)
    parser.add_argument('-r','--recurse',help="Process recursively", action='store_true', default=False)
    parser.add_argument('-a','--autodeclare',help="Attempt to automatically declare missing annotations", action='store_true', default=False)
    parser.add_argument('-q','--quick',help="Quicker validation; skips RelaxNG validation and serialisation checks", action='store_true', default=False)
    parser.add_argument('-E','--extension', type=str,help="Extension", action='store',default="xml")
    parser.add_argument('-W','--nowarn',help="Suppress warnings", action='store_true', default=False)
    parser.add_argument('-i','--ignore',help="Always report a successful exit code, even in case of validation errors", action='store_true', default=False)
    parser.add_argument('-t','--stricttextvalidation',help="Treat text validation errors strictly for FoLiA < v1.5,  it always enabled for for FoLiA v1.5+ regardless of this parameter", action='store_true', default=False)
    parser.add_argument('-o','--output',help="Output document to stdout. This output contains added proof of validation to the provenance chain (not cryptographically secure though!)", action='store_true', default=False)
    parser.add_argument('-k','--keepversion',help="Attempt to keep an older FoLiA version (not always guaranteed to work)", action='store_true', default=False)
    parser.add_argument('-D','--debug',type=int,help="Debug level", action='store',default=0)
    parser.add_argument('-b','--traceback',help="Provide a full traceback on validation errors", action='store_true', default=False)
    parser.add_argument('-x','--explicit',help="Serialise to explicit form, this generates more verbose XML and simplified the job for parsers as implicit information is made explicit", action='store_true', default=False)
    parser.add_argument('--fixunassignedprocessor',help="Fixes invalid FoLiA that does not explicitly assign a processor to an annotation when multiple processors are possible (and there is therefore no default). The first processor will be used in this case.", action='store_true', default=False)
    return parser

def main():
    parser = argparse.ArgumentParser(description=__doc__, formatter_class=argparse.ArgumentDefaultsHelpFormatter)
    parser.add_argument('-v','-V','--version',help="Show version information", action='version', version="FoLiA-tools v" + TOOLVERSION + ", using FoLiA v" + folia.FOLIAVERSION + " with library FoLiApy v" + folia.LIBVERSION, default=False)
    commandparser(parser)
    parser.add_argument('files', nargs='*', help='Files (and/or directories) to validate')
    args = parser.parse_args()

    schema  = lxml.etree.RelaxNG(folia.relaxng())

    if args.explicit:
        args.output = True

    if args.files:
        success = True
        skipnext = False
        for file in args.files:
            r = False
            if os.path.isdir(file):
                r = processdir(file,schema, **args.__dict__)
            elif os.path.isfile(file):
                r = validate(file, schema, **args.__dict__)
            else:
                print("ERROR: File or directory not found: " + file,file=sys.stderr)
                sys.exit(3)
            if not r: success= False
        if not success and not args.ignore:
            sys.exit(1)
    else:
        print("ERROR: No files specified. Add --help for usage details.",file=sys.stderr)
        sys.exit(2)

if __name__ == "__main__":
    main()



#!/usr/bin/env python
# -*- coding: utf8 -*-

"""Merges annotations from two or more FoLiA documents. Structural elements are never added. Annotations can only be merged if their parent elements have IDs."""

import getopt
import sys
from foliatools import VERSION
import folia.main as folia

def usage():
    print("foliamerge",file=sys.stderr)
    print("  by Maarten van Gompel (proycon)",file=sys.stderr)
    print("  Radboud University Nijmegen",file=sys.stderr)
    print("  2017-2019 - Licensed under GPLv3",file=sys.stderr)
    print("",file=sys.stderr)
    print(__doc__,file=sys.stderr)
    print("",file=sys.stderr)
    print("Usage: foliamerge [options] file1 file2 file3 ... ",file=sys.stderr)
    print("",file=sys.stderr)
    print("Options:",file=sys.stderr)
    print("  -o [file]                    Output file",file=sys.stderr)
    print("  -s                           Substitute: use first input file as output as well",file=sys.stderr)
    print("  -a                           Annotations from file2 onwards are included as alternatives",file=sys.stderr)

def attach(parent, child):
    """Rather than copy children, we just abduct them from the other-doc parent"""
    child.doc = parent.doc
    child.parent = parent
    child.setparents()

def reID(doc, element, oldprefix, newprefix):
    if element.id:
        element.id = element.id.replace(oldprefix + '.', newprefix + '.')
    for child in element.data:
        if isinstance(child, folia.AbstractElement) and not isinstance(child, folia.AbstractStructureElement):
            reID(doc, child, oldprefix, newprefix)

def mergechildren(parent, outputdoc, asalternative):
    if hasattr(parent,'merged'): return 0
    merges = 0
    for e in parent:
        if isinstance(e, (folia.AbstractInlineAnnotation, folia.AbstractAnnotationLayer)) and parent.id and not hasattr(e, 'merged'):
            try:
                e.ANNOTATIONTYPE
            except:
                continue

            if (e.ANNOTATIONTYPE, e.set) in outputdoc.annotations:
                assert e.parent == parent
                try:
                    newparent = outputdoc[parent.id]
                except:
                    continue #parent does not exist, nothing to merge here
                if isinstance(newparent, (folia.Alternative, folia.AlternativeLayers)):
                    #we do not merge alternatives, each alternative has its own scope
                    continue
                #check if the annotation already exists
                #print("DEBUG: Adding annotation type " + e.__class__.__name__ + ", set " + e.set + ", under " + newparent.id, file=sys.stderr)
                if asalternative:
                    if isinstance(e, folia.AbstractInlineAnnotation):
                        print("Adding Annotation type " + e.__class__.__name__ + ", set " + str(e.set) + " to " + newparent.id + " as alternative", file=sys.stderr)
                        alt = newparent.append(folia.Alternative, generate_id_in=newparent)
                        reID(newparent.doc, e, newparent.id, alt.id)
                        alt.append(e)
                        e.setdoc(outputdoc)
                        assert e.parent is alt
                        alt.merged = True
                        e.merged = True
                        merges += 1
                    elif isinstance(e, folia.AbstractAnnotationLayer):
                        print("Adding Annotation type " + e.__class__.__name__ + ", set " + str(e.set) + " to " + newparent.id + " as alternative", file=sys.stderr)
                        alt = newparent.append(folia.AlternativeLayers, generate_id_in=newparent)
                        reID(newparent.doc, e, newparent.id, alt.id)
                        alt.append(e)
                        assert e.parent is alt
                        e.setdoc(outputdoc)
                        alt.merged = True
                        e.merged = True
                        merges += 1
                elif isinstance(e, folia.AbstractInlineAnnotation) and newparent.hasannotation(e.__class__, e.set):
                    print("Annotation type " + e.__class__.__name__ + ", set " + e.set + ", under " + newparent.id + " , already exists... skipping", file=sys.stderr)
                    pass
                elif isinstance(e, folia.AbstractAnnotationLayer) and newparent.hasannotationlayer(e.__class__, e.set):
                    print("Annotation type " + e.__class__.__name__ + ", set " + e.set + ", under " + newparent.id + " , already exists... skipping", file=sys.stderr)
                    pass
                else:
                    print("Adding Annotation type " + e.__class__.__name__ + ", set " + str(e.set) + " to " + newparent.id, file=sys.stderr)
                    newparent.append(e)
                    assert e.parent is newparent
                    e.setdoc(outputdoc)
                    e.merged = True
                    merges += 1
        elif isinstance(e, folia.AbstractElement):
            merges += mergechildren(e, outputdoc, asalternative)
    return merges



def foliamerge(outputfile, *files, **kwargs):
    asalternative = 'asalternative' in kwargs and kwargs['asalternative']
    outputdoc = None
    merges = 0

    for i, filename in enumerate(files):
        print("Processing " + filename, file=sys.stderr)
        inputdoc = folia.Document(file=filename)
        if i == 0:
             print("(pivot document)",file=sys.stderr)
             outputdoc = inputdoc
        else:
            print("(merging document)",file=sys.stderr)

            for annotationtype,set in inputdoc.annotations:
                if not outputdoc.declared(annotationtype,set):
                    outputdoc.declare( annotationtype, set)

            for e in inputdoc:
                merges += mergechildren(e, outputdoc, asalternative)

    if outputfile and merges > 0:
        outputdoc.provenance.append(folia.Processor.create(name="foliamerge", version=VERSION))
        outputdoc.save(outputfile)

    return outputdoc

def main():
    try:
        opts, args = getopt.getopt(sys.argv[1:], "o:sha", ["help"])
    except getopt.GetoptError as err:
        print(str(err),file=sys.stderr)
        usage()
        sys.exit(2)

    outputfile = None
    substitute = False
    asalternative = False

    for o, a in opts:
        if o == '-h' or o == '--help':
            usage()
            sys.exit(0)
        elif o == '-o':
            outputfile = a
        elif o == '-s':
            substitute = True
        elif o == '-a':
            asalternative = True
        else:
            raise Exception("No such option: " + o)

    if len(args) < 2:
        print("ERROR: At least two files need to be specified",file=sys.stderr)
        sys.exit(2)

    if substitute:
        outputfile = args[0]

    outputdoc = foliamerge(outputfile, *args, asalternative=asalternative)
    if not outputfile:
        xml = outputdoc.xmlstring()
        print(xml)


if __name__ == "__main__":
    main()






#!/usr/bin/env python
# -*- coding: utf8 -*-

"""
Splits a FoLiA document into multiple separate FoLiA documents based on user-specified selection criteria.
"""

import sys
import os
import argparse
from copy import deepcopy
from foliatools import VERSION as TOOLVERSION
import folia.main as folia
import folia.fql as fql

def split(doc, expression, batchsize=1, copymetadata=False, require_submetadata=False,  suffix_template="_{:04d}", alter_ids=False, external=False, callback=None, deep=False):
    query = fql.Query("SELECT " + expression)
    childdoc = None
    prevmatch = None
    docnr = 0
    substituted = False #ensure we only substitute once per batch (when external is enabled)
    if external:
        proc = doc.provenance.append(folia.Processor.create(name="foliasplit", version=TOOLVERSION))
    for match in query(doc):
        if require_submetadata and match.metadata is None:
            continue
        if prevmatch and prevmatch in match.ancestors():
            #prevent recursion within elements that are already split into documents
            continue
        if callback and not callback(match):
            continue
        if not childdoc:
            docnr += 1
            id_suffix = suffix_template.format(docnr)
            childdoc = folia.Document(id=doc.id + id_suffix )
            if copymetadata:
                childdoc.metadata = deepcopy(doc.metadata)
            if deep:
                childdoc.annotations = deepcopy(doc.annotations)
                childdoc.annotators = deepcopy(doc.annotators)
                childdoc.annotator2processor_map = deepcopy(doc.annotator2processor_map)
                childdoc.groupannotations = deepcopy(doc.groupannotations)
                childdoc.alias_set = deepcopy(doc.alias_set)
                childdoc.set_alias = deepcopy(doc.set_alias)
                childdoc.textclasses = deepcopy(doc.textclasses)
            else:
                childdoc.annotations = doc.annotations
                childdoc.annotators = doc.annotators
                childdoc.annotator2processor_map = doc.annotator2processor_map
                childdoc.groupannotations = doc.groupannotations
                childdoc.alias_set = doc.alias_set
                childdoc.set_alias = doc.set_alias
                childdoc.textclasses = doc.textclasses
            childdoc.provenance = deepcopy(doc.provenance)
            if not external:
                childdoc.provenance.append(folia.Processor.create(name="foliasplit", version=TOOLVERSION))
            if match.metadata:
                #copy submetadata to new document's metadata
                try:
                    submetadata = doc.submetadata[match.metadata]
                except KeyError:
                    print("WARNING: Submetadata not found: " + match.metadata,file=sys.stderr)
                    submetadata = None
                if submetadata:
                    for key, value in submetadata.items():
                        childdoc.metadata[key] = value
            #add main body element (text or speech)
            bodyid = doc.id + id_suffix + "."  + doc.data[0].XMLTAG
            if bodyid in doc:
                bodyid = bodyid + ".split"
            body = childdoc.append(doc.data[0].__class__(childdoc, id=bodyid))
            if external:
                if substituted:
                    match.parent.data.remove(match)
                else:
                    i = match.parent.getindex(match)
                    if i == -1:
                        raise Exception("Unable to find child from parent! This shouldn't happen")
                    match.parent.data[i] = folia.External(doc, src=childdoc.id + ".folia.xml", id=childdoc.id, processor=proc)
                    substituted = True
        if deep:
            matchcopy = match.copy(childdoc, id_suffix if alter_ids else "") #deep copy
        else:
            matchcopy = match.move(childdoc, id_suffix if alter_ids else "") #shallow copy
        matchcopy.metadata = None
        body.append(matchcopy)
        if len(body.data) == batchsize:
            yield childdoc
            childdoc = None
            body = None
            substituted = False
        prevmatch = match
    if childdoc:
        yield childdoc



def main():
    parser = argparse.ArgumentParser(description=__doc__, formatter_class=argparse.ArgumentDefaultsHelpFormatter)
    parser.add_argument('-v','-V','--version',help="Show version information", action='version', version="FoLiA-tools v" + TOOLVERSION + ", using FoLiA v" + folia.FOLIAVERSION + " with library FoLiApy v" + folia.LIBVERSION, default=False)
    parser.add_argument('--copymetadata','-m',help="Copy all metadata from the parent document to the children", action='store_true', required=False)
    parser.add_argument('--batchsize','-b',type=int, help="Batch size: create documents with this many matches", action='store', required=False, default=1)
    parser.add_argument('--alterids','-i',help="Alter the IDs of all split elements by appending a suffix", action='store_true', required=False)
    parser.add_argument('--deep',help="Make a deep copy (slower)", action='store_true', required=False)
    parser.add_argument('--suffixtemplate',help="A template for adding suffixes to IDs, in Python's format syntax", action='store_true', required=False, default="_{:04d}")
    parser.add_argument('--submetadata',help="Only split elements that have associated submetadata (extra parameter as the query can't capture this)", action='store_true', required=False)
    parser.add_argument('--query','-q',type=str, help="Query to select elements to split, this is an FQL SELECT expression without the SELECT statement, it can be as simple as the element type, e.g. s for sentences or more complex like: 'div OF https://your.set WHERE class = \"chapter\"' ", action='store', required=True)
    parser.add_argument('--outputdir','-o',type=str, help="Output directory", action='store', required=False,default=".")
    parser.add_argument("--external","-x",  action='store_true', help="Replace split-off elements with <external> elements in the original input document. Output the input document (this may overwrite the input if you it's in your immediate working dir and you didn't specify a custom --outputdir!)" , required=False)
    parser.add_argument("files", nargs="*", help="The FoLiA documents to split")
    args = parser.parse_args()

    if args.external and args.batchsize > 1:
        print("WARNING: You are using the external mechanism with a batchsize greater than 1, this may produce wrongly ordered output depending on your query and your input. Please inspect your parent document to make sure the results are sensible.",file=sys.stderr)

    if len(args.files) < 1:
        print("No files specified. Run with --help for usage info.", file=sys.stderr)

    #parse query prior to reading files just to make sure there are no syntax errors
    fql.Query("SELECT " + args.query)

    for filename in args.files:
        print("Loading " + filename, file=sys.stderr)
        doc = folia.Document(file=filename)
        for i, childdoc in enumerate(split(doc, args.query, args.batchsize, args.copymetadata, args.submetadata, args.suffixtemplate, args.alterids, args.external, None, args.deep)):
            print("#" + str(i+1) + " - " + childdoc.id + ".folia.xml", file=sys.stderr)
            childdoc.save(os.path.join(args.outputdir, childdoc.id) + ".folia.xml")
            del childdoc

        if args.external:
            print("Saving parent document - " + os.path.basename(filename), file=sys.stderr)
            doc.save(os.path.join(args.outputdir, os.path.basename(filename)))


if __name__ == "__main__":
    main()
